from collections.abc import Callable

import pytest

from hugr import tys as ht

from hugr import Wire

from guppylang.decorator import guppy
from guppylang.definition.custom import CustomCallCompiler
from guppylang.module import GuppyModule
from guppylang.std.builtins import array
from guppylang.std.quantum import qubit

import guppylang.std.quantum as quantum


def test_id(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy(module)
    def main(x: int) -> int:
        return foo(x)

    validate(module.compile())


def test_id_nested(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy(module)
    def main(x: int) -> int:
        return foo(foo(foo(x)))

    validate(module.compile())


def test_use_twice(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy(module)
    def main(x: int, y: bool) -> None:
        foo(x)
        foo(y)

    validate(module.compile())


def test_define_twice(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy.declare(module)
    def bar(x: T) -> T:  # Reuse same type var!
        ...

    @guppy(module)
    def main(x: bool, y: float) -> None:
        foo(x)
        foo(y)

    validate(module.compile())


def test_return_tuple_implicit(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy(module)
    def main(x: int) -> tuple[int, int]:
        return foo((x, 0))

    validate(module.compile())


def test_same_args(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(x: T, y: T) -> None: ...

    @guppy(module)
    def main(x: int) -> None:
        foo(x, 42)

    validate(module.compile())


def test_different_args(validate):
    module = GuppyModule("test")
    S = guppy.type_var("S", module=module)
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(x: S, y: T, z: tuple[S, T]) -> T: ...

    @guppy(module)
    def main(x: int, y: float) -> float:
        return foo(x, y, (x, y)) + foo(y, 42.0, (0.0, y))

    validate(module.compile())


def test_nat_args(validate):
    module = GuppyModule("test")
    n = guppy.nat_var("n", module=module)

    @guppy.declare(module)
    def foo(x: array[int, n]) -> array[int, n]: ...

    @guppy(module)
    def main(x: array[int, 42]) -> array[int, 42]:
        return foo(x)


def test_infer_basic(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo() -> T: ...

    @guppy(module)
    def main() -> None:
        x: int = foo()

    validate(module.compile())


def test_infer_list(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo() -> T: ...

    @guppy(module)
    def main() -> None:
        xs: list[int] = [foo()]
        ys = [1.0, foo()]

    validate(module.compile())


def test_infer_nested(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo() -> T: ...

    @guppy.declare(module)
    def bar(x: T) -> T: ...

    @guppy(module)
    def main() -> None:
        x: int = bar(foo())

    validate(module.compile())


def test_infer_left_to_right(validate):
    module = GuppyModule("test")
    S = guppy.type_var("S", module=module)
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo() -> T: ...

    @guppy.declare(module)
    def bar(x: T, y: T, z: S, a: tuple[T, S]) -> None: ...

    @guppy(module)
    def main() -> None:
        bar(42, foo(), False, foo())

    validate(module.compile())


def test_type_apply_basic(validate):
    module = GuppyModule("test")
    S = guppy.type_var("S", module=module)
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy.declare(module)
    def bar(x: S, y: T) -> S: ...

    @guppy(module)
    def main() -> tuple[int, float, float]:
        x = foo[int](0)
        y = foo[float](1.0)
        z = bar[float, int](y, x)
        return x, y, z

    validate(module.compile())


def test_type_apply_higher_order(validate):
    module = GuppyModule("test")
    S = guppy.type_var("S", module=module)
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy.declare(module)
    def bar(x: S, y: T) -> S: ...

    @guppy(module)
    def main() -> tuple[int, float, float]:
        f = foo[int]
        g = foo[float]
        h = bar[float, int]
        x = f(0)
        y = g(1.0)
        z = h(y, x)
        return x, y, z

    validate(module.compile())


def test_type_apply_nat(validate):
    module = GuppyModule("test")
    n = guppy.nat_var("n", module=module)

    @guppy.declare(module)
    def foo(x: array[int, n]) -> int: ...

    @guppy(module)
    def main() -> int:
        return foo[0](array()) + foo[2](array(1, 2))

    validate(module.compile())


def test_type_apply_empty_tuple(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(x: T) -> None:
        ...

    @guppy(module)
    def main() -> None:
        # `()` is the type of an empty tuple (`tuple[]` is not syntactically valid)
        foo[()]

    validate(module.compile())


def test_pass_poly_basic(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(f: Callable[[T], T]) -> None: ...

    @guppy.declare(module)
    def bar(x: int) -> int: ...

    @guppy(module)
    def main() -> None:
        foo(bar)

    validate(module.compile())


def test_pass_poly_cross(validate):
    module = GuppyModule("test")
    S = guppy.type_var("S", module=module)
    T = guppy.type_var("T", module=module)

    @guppy.declare(module)
    def foo(f: Callable[[S], int]) -> None: ...

    @guppy.declare(module)
    def bar(x: bool) -> T: ...

    @guppy(module)
    def main() -> None:
        foo(bar)

    validate(module.compile())


def test_linear(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    T = guppy.type_var("T", copyable=False, droppable=False, module=module)

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy(module)
    def main(q: qubit) -> qubit:
        return foo(q)

    validate(module.compile())


def test_affine(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", copyable=False, droppable=True, module=module)

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy(module)
    def main(a: array[int, 7]) -> None:
        foo(a)

    validate(module.compile())


def test_relevant(validate):
    module = GuppyModule("test")
    T = guppy.type_var("T", copyable=True, droppable=False, module=module)

    @guppy.type(ht.Bool, copyable=True, droppable=False, module=module)
    class R: ...

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy(module)
    def main(r: R) -> R:
        r_copy = r
        return foo(r_copy)

    validate(module.compile())


def test_pass_nonlinear(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    T = guppy.type_var("T", copyable=False, droppable=False, module=module)

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy(module)
    def main(x: int) -> None:
        foo(x)

    validate(module.compile())


def test_pass_linear(validate):
    module = GuppyModule("test")
    module.load_all(quantum)
    T = guppy.type_var("T", copyable=False, droppable=False, module=module)

    @guppy.declare(module)
    def foo(f: Callable[[T], T]) -> None: ...

    @guppy.declare(module)
    def bar(q: qubit) -> qubit: ...

    @guppy(module)
    def main() -> None:
        foo(bar)

    validate(module.compile())


def test_custom_higher_order():
    class CustomCompiler(CustomCallCompiler):
        def compile(self, args: list[Wire]) -> list[Wire]:
            return args

    module = GuppyModule("test")
    T = guppy.type_var("T", module=module)

    @guppy.custom(module, CustomCompiler())
    def foo(x: T) -> T: ...

    @guppy(module)
    def main(x: int) -> int:
        f: Callable[[int], int] = foo
        return f(x)


@pytest.mark.skip("Higher-order polymorphic functions are not yet supported")
def test_higher_order_value(validate):
    module = GuppyModule("test")
    T = guppy.type_var(module, "T")

    @guppy.declare(module)
    def foo(x: T) -> T: ...

    @guppy.declare(module)
    def bar(x: T) -> T: ...

    @guppy(module)
    def main(b: bool) -> int:
        f = foo if b else bar
        return f(42)

    validate(module.compile())



"""Dummy module used in `test_imports.py`"""

from guppylang import GuppyModule, guppy
from guppylang.std._internal.util import unsupported_op

mod_b = GuppyModule("mod_b")


@guppy(mod_b)
def f(x: bool) -> bool:
    return not x


@guppy.hugr_op(unsupported_op("h"), module=mod_b)
def h() -> int: ...


@guppy.struct(mod_b)
class MyType:
    x: int

    @guppy(mod_b)
    def __pos__(self: "MyType") -> "MyType":
        return self



"""Compilers building array functions on top of hugr standard operations."""

from __future__ import annotations

from typing import TYPE_CHECKING, Final

from hugr import Wire, ops
from hugr import tys as ht
from hugr.std.collections.array import EXTENSION

from guppylang.compiler.core import (
    GlobalConstId,
)
from guppylang.definition.custom import CustomCallCompiler
from guppylang.definition.value import CallReturnWires
from guppylang.error import InternalGuppyError
from guppylang.std._internal.compiler.arithmetic import convert_itousize
from guppylang.std._internal.compiler.prelude import (
    build_expect_none,
    build_unwrap,
    build_unwrap_left,
    build_unwrap_right,
)
from guppylang.tys.arg import ConstArg, TypeArg
from guppylang.tys.builtin import int_type

if TYPE_CHECKING:
    from hugr.build import function as hf

# ------------------------------------------------------
# --------------- std.array operations -----------------
# ------------------------------------------------------


def _instantiate_array_op(
    name: str,
    elem_ty: ht.Type,
    length: ht.TypeArg,
    inp: list[ht.Type],
    out: list[ht.Type],
) -> ops.ExtOp:
    return EXTENSION.get_op(name).instantiate(
        [length, ht.TypeTypeArg(elem_ty)], ht.FunctionType(inp, out)
    )


def array_type(elem_ty: ht.Type, length: ht.TypeArg) -> ht.ExtType:
    """Returns the hugr type of a fixed length array."""
    elem_arg = ht.TypeTypeArg(elem_ty)
    return EXTENSION.types["array"].instantiate([length, elem_arg])


def array_new(elem_ty: ht.Type, length: int) -> ops.ExtOp:
    """Returns an operation that creates a new fixed length array."""
    length_arg = ht.BoundedNatArg(length)
    arr_ty = array_type(elem_ty, length_arg)
    return _instantiate_array_op(
        "new_array", elem_ty, length_arg, [elem_ty] * length, [arr_ty]
    )


def array_get(elem_ty: ht.Type, length: ht.TypeArg) -> ops.ExtOp:
    """Returns an array `get` operation."""
    assert elem_ty.type_bound() == ht.TypeBound.Copyable
    arr_ty = array_type(elem_ty, length)
    return _instantiate_array_op(
        "get", elem_ty, length, [arr_ty, ht.USize()], [ht.Option(elem_ty)]
    )


def array_set(elem_ty: ht.Type, length: ht.TypeArg) -> ops.ExtOp:
    """Returns an array `set` operation."""
    arr_ty = array_type(elem_ty, length)
    return _instantiate_array_op(
        "set",
        elem_ty,
        length,
        [arr_ty, ht.USize(), elem_ty],
        [ht.Either([elem_ty, arr_ty], [elem_ty, arr_ty])],
    )


def array_pop(elem_ty: ht.Type, length: int, from_left: bool) -> ops.ExtOp:
    """Returns an operation that pops an element from the left of an array."""
    assert length > 0
    length_arg = ht.BoundedNatArg(length)
    arr_ty = array_type(elem_ty, length_arg)
    popped_arr_ty = array_type(elem_ty, ht.BoundedNatArg(length - 1))
    op = "pop_left" if from_left else "pop_right"
    return _instantiate_array_op(
        op, elem_ty, length_arg, [arr_ty], [ht.Option(elem_ty, popped_arr_ty)]
    )


def array_discard_empty(elem_ty: ht.Type) -> ops.ExtOp:
    """Returns an operation that discards an array of length zero."""
    arr_ty = array_type(elem_ty, ht.BoundedNatArg(0))
    return EXTENSION.get_op("discard_empty").instantiate(
        [ht.TypeTypeArg(elem_ty)], ht.FunctionType([arr_ty], [])
    )


def array_scan(
    elem_ty: ht.Type,
    length: ht.TypeArg,
    new_elem_ty: ht.Type,
    accumulators: list[ht.Type],
) -> ops.ExtOp:
    """Returns an operation that maps and folds a function across an array."""
    ty_args = [
        length,
        ht.TypeTypeArg(elem_ty),
        ht.TypeTypeArg(new_elem_ty),
        ht.SequenceArg([ht.TypeTypeArg(acc) for acc in accumulators]),
        ht.ExtensionsArg([]),
    ]
    ins = [
        array_type(elem_ty, length),
        ht.FunctionType([elem_ty, *accumulators], [new_elem_ty, *accumulators]),
        *accumulators,
    ]
    outs = [array_type(new_elem_ty, length), *accumulators]
    return EXTENSION.get_op("scan").instantiate(ty_args, ht.FunctionType(ins, outs))


def array_map(elem_ty: ht.Type, length: ht.TypeArg, new_elem_ty: ht.Type) -> ops.ExtOp:
    """Returns an operation that maps a function across an array."""
    return array_scan(elem_ty, length, new_elem_ty, accumulators=[])


def array_repeat(elem_ty: ht.Type, length: ht.TypeArg) -> ops.ExtOp:
    """Returns an array `repeat` operation."""
    return EXTENSION.get_op("repeat").instantiate(
        [length, ht.TypeTypeArg(elem_ty), ht.ExtensionsArg([])],
        ht.FunctionType(
            [ht.FunctionType([], [elem_ty])], [array_type(elem_ty, length)]
        ),
    )


# ------------------------------------------------------
# --------- Custom compilers for non-native ops --------
# ------------------------------------------------------


class ArrayCompiler(CustomCallCompiler):
    """Base class for custom array op compilers."""

    @property
    def elem_ty(self) -> ht.Type:
        """The element type for the array op that is being compiled."""
        match self.type_args:
            case [TypeArg(ty=elem_ty), _]:
                return elem_ty.to_hugr()
            case _:
                raise InternalGuppyError("Invalid array type args")

    @property
    def length(self) -> ht.TypeArg:
        """The length for the array op that is being compiled."""
        match self.type_args:
            case [_, ConstArg(const)]:  # Const includes both literals and variables
                return const.to_arg().to_hugr()
            case _:
                raise InternalGuppyError("Invalid array type args")


class NewArrayCompiler(ArrayCompiler):
    """Compiler for the `array.__new__` function."""

    def build_classical_array(self, elems: list[Wire]) -> Wire:
        """Lowers a call to `array.__new__` for classical arrays."""
        # See https://github.com/CQCL/guppylang/issues/629
        return self.build_linear_array(elems)

    def build_linear_array(self, elems: list[Wire]) -> Wire:
        """Lowers a call to `array.__new__` for linear arrays."""
        elem_opts = [
            self.builder.add_op(ops.Some(self.elem_ty), elem) for elem in elems
        ]
        return self.builder.add_op(
            array_new(ht.Option(self.elem_ty), len(elems)), *elem_opts
        )

    def compile(self, args: list[Wire]) -> list[Wire]:
        if self.elem_ty.type_bound() == ht.TypeBound.Any:
            return [self.build_linear_array(args)]
        else:
            return [self.build_classical_array(args)]


ARRAY_GETITEM_CLASSICAL: Final[GlobalConstId] = GlobalConstId.fresh(
    "array.__getitem__.classical"
)
ARRAY_GETITEM_LINEAR: Final[GlobalConstId] = GlobalConstId.fresh(
    "array.__getitem__.linear"
)
ARRAY_SETITEM_CLASSICAL: Final[GlobalConstId] = GlobalConstId.fresh(
    "array.__setitem__.classical"
)
ARRAY_SETITEM_LINEAR: Final[GlobalConstId] = GlobalConstId.fresh(
    "array.__setitem__.linear"
)
ARRAY_ITER_ASSERT_ALL_USED_HELPER: Final[GlobalConstId] = GlobalConstId.fresh(
    "ArrayIter._assert_all_used.helper"
)


class ArrayGetitemCompiler(ArrayCompiler):
    """Compiler for the `array.__getitem__` function."""

    def _getitem_ty(self, bound: ht.TypeBound) -> ht.PolyFuncType:
        """Constructs a polymorphic function type for `__getitem__`"""
        # a(Option(T), N), int -> T, a(Option(T), N)
        # Array element type parameter
        elem_ty_param = ht.TypeTypeParam(bound)
        # Array length parameter
        length_param = ht.BoundedNatParam()
        return ht.PolyFuncType(
            params=[elem_ty_param, length_param],
            body=ht.FunctionType(
                input=[
                    array_type(
                        ht.Option(ht.Variable(0, bound)),
                        ht.VariableArg(1, length_param),
                    ),
                    int_type().to_hugr(),
                ],
                output=[
                    ht.Variable(0, bound),
                    array_type(
                        ht.Option(ht.Variable(0, bound)),
                        ht.VariableArg(1, length_param),
                    ),
                ],
            ),
        )

    def _build_classical_getitem(self, func: hf.Function) -> None:
        """Constructs a generic function for `__getitem__` for classical arrays."""
        elem_ty = ht.Variable(0, ht.TypeBound.Copyable)
        length = ht.VariableArg(1, ht.BoundedNatParam())

        # See https://github.com/CQCL/guppylang/issues/629
        elem_opt_ty = ht.Option(elem_ty)
        idx = func.add_op(convert_itousize(), func.inputs()[1])
        result = func.add_op(
            array_get(elem_opt_ty, length),
            func.inputs()[0],
            idx,
        )
        elem_opt = build_unwrap(func, result, "Array index out of bounds")
        elem = build_unwrap(func, elem_opt, "array.__getitem__: Internal error")

        # Return input array unmodified for consistency with linear implementation.
        func.set_outputs(elem, func.inputs()[0])

    def _build_linear_getitem(self, func: hf.Function) -> None:
        """Constructs function to call `array.__getitem__` for linear arrays."""
        elem_ty = ht.Variable(0, ht.TypeBound.Any)
        length = ht.VariableArg(1, ht.BoundedNatParam())

        elem_opt_ty = ht.Option(elem_ty)
        none = func.add_op(ops.Tag(0, elem_opt_ty))
        idx = func.add_op(convert_itousize(), func.inputs()[1])
        result = func.add_op(
            array_set(elem_opt_ty, length),
            func.inputs()[0],
            idx,
            none,
        )
        elem_opt, array = build_unwrap_right(func, result, "Array index out of bounds")
        elem = build_unwrap(
            func, elem_opt, "Linear array element has already been used"
        )

        func.set_outputs(elem, array)

    def _build_call_getitem(
        self,
        func: hf.Function,
        array: Wire,
        idx: Wire,
    ) -> CallReturnWires:
        """Inserts a call to `array.__getitem__`."""
        concrete_func_ty = ht.FunctionType(
            input=[
                array_type(ht.Option(self.elem_ty), self.length),
                int_type().to_hugr(),
            ],
            output=[self.elem_ty, array_type(ht.Option(self.elem_ty), self.length)],
        )
        type_args = [ht.TypeTypeArg(self.elem_ty), self.length]
        func_call = self.builder.call(
            func.parent_node,
            array,
            idx,
            instantiation=concrete_func_ty,
            type_args=type_args,
        )
        outputs = list(func_call.outputs())
        return CallReturnWires(
            regular_returns=[outputs[0]],
            inout_returns=[outputs[1]],
        )

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [array, idx] = args
        if self.elem_ty.type_bound() == ht.TypeBound.Any:
            func_ty = self._getitem_ty(ht.TypeBound.Any)
            func, already_exists = self.ctx.declare_global_func(
                ARRAY_GETITEM_LINEAR, func_ty
            )
            if not already_exists:
                self._build_linear_getitem(func)
        else:
            func_ty = self._getitem_ty(ht.TypeBound.Copyable)
            func, already_exists = self.ctx.declare_global_func(
                ARRAY_GETITEM_CLASSICAL, func_ty
            )
            if not already_exists:
                self._build_classical_getitem(func)
        return self._build_call_getitem(
            func=func,
            array=array,
            idx=idx,
        )

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


class ArraySetitemCompiler(ArrayCompiler):
    """Compiler for the `array.__setitem__` function."""

    def _setitem_ty(self, bound: ht.TypeBound) -> ht.PolyFuncType:
        """Constructs a polymorphic function type for `__setitem__`"""
        # a(Option(T), N), int, T -> a(Option(T), N)
        elem_ty_param = ht.TypeTypeParam(bound)
        length_param = ht.BoundedNatParam()
        return ht.PolyFuncType(
            params=[elem_ty_param, length_param],
            body=ht.FunctionType(
                input=[
                    array_type(
                        ht.Option(ht.Variable(0, bound)),
                        ht.VariableArg(1, length_param),
                    ),
                    int_type().to_hugr(),
                    ht.Variable(0, bound),
                ],
                output=[
                    array_type(
                        ht.Option(ht.Variable(0, bound)),
                        ht.VariableArg(1, length_param),
                    ),
                ],
            ),
        )

    def _build_classical_setitem(self, func: hf.Function) -> None:
        """Constructs a generic function for `__setitem__` for classical arrays."""
        elem_ty = ht.Variable(0, ht.TypeBound.Copyable)
        length = ht.VariableArg(1, ht.BoundedNatParam())

        elem_opt_ty = ht.Option(elem_ty)
        idx = func.add_op(convert_itousize(), func.inputs()[1])
        elem_opt = func.add_op(ops.Some(elem_ty), func.inputs()[2])
        result = func.add_op(
            array_set(elem_opt_ty, length),
            func.inputs()[0],
            idx,
            elem_opt,
        )
        _, array = build_unwrap_right(func, result, "Array index out of bounds")

        func.set_outputs(array)

    def _build_linear_setitem(self, func: hf.Function) -> None:
        """Constructs function to call `array.__setitem__` for linear arrays."""
        elem_ty = ht.Variable(0, ht.TypeBound.Any)
        length = ht.VariableArg(1, ht.BoundedNatParam())

        elem_opt_ty = ht.Option(elem_ty)
        elem = func.add_op(ops.Some(elem_ty), func.inputs()[2])
        idx = func.add_op(convert_itousize(), func.inputs()[1])
        result = func.add_op(
            array_set(elem_opt_ty, length),
            func.inputs()[0],
            idx,
            elem,
        )
        old_elem_opt, array = build_unwrap_right(
            func, result, "Array index out of bounds"
        )
        build_unwrap_left(func, old_elem_opt, "Linear array element has not been used")

        func.set_outputs(array)

    def _build_call_setitem(
        self,
        func: hf.Function,
        array: Wire,
        idx: Wire,
        elem: Wire,
    ) -> CallReturnWires:
        """Inserts a call to `array.__setitem__`."""
        concrete_func_ty = ht.FunctionType(
            input=[
                array_type(ht.Option(self.elem_ty), self.length),
                int_type().to_hugr(),
                self.elem_ty,
            ],
            output=[array_type(ht.Option(self.elem_ty), self.length)],
        )
        type_args = [ht.TypeTypeArg(self.elem_ty), self.length]
        func_call = self.builder.call(
            func.parent_node,
            array,
            idx,
            elem,
            instantiation=concrete_func_ty,
            type_args=type_args,
        )
        return CallReturnWires(
            regular_returns=[],
            inout_returns=list(func_call.outputs()),
        )

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [array, idx, elem] = args
        if self.elem_ty.type_bound() == ht.TypeBound.Any:
            func_ty = self._setitem_ty(ht.TypeBound.Any)
            func, already_exists = self.ctx.declare_global_func(
                ARRAY_SETITEM_LINEAR, func_ty
            )
            if not already_exists:
                self._build_linear_setitem(func)
        else:
            func_ty = self._setitem_ty(ht.TypeBound.Copyable)
            func, already_exists = self.ctx.declare_global_func(
                ARRAY_SETITEM_CLASSICAL, func_ty
            )
            if not already_exists:
                self._build_classical_setitem(func)
        return self._build_call_setitem(func=func, array=array, idx=idx, elem=elem)

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Call compile_with_inouts instead")


class ArrayIterAsertAllUsedCompiler(ArrayCompiler):
    """Compiler for the `ArrayIter._assert_all_used` method."""

    def compile(self, args: list[Wire]) -> list[Wire]:
        # For linear array iterators, map the array of optional elements to an
        # `array[None, n]` that we can discard.
        if self.elem_ty.type_bound() == ht.TypeBound.Any:
            elem_opt_ty = ht.Option(self.elem_ty)
            unit_ty = ht.UnitSum(1)
            # Instantiate `unwrap_none` function
            func = self.builder.load_function(
                self.define_unwrap_none_helper(),
                type_args=[ht.TypeTypeArg(self.elem_ty)],
                instantiation=ht.FunctionType([elem_opt_ty], [unit_ty]),
            )
            # Map it over the array so that the resulting array is no longer linear and
            # can be discarded
            [array_iter] = args
            array, _ = self.builder.add_op(ops.UnpackTuple(), array_iter)
            self.builder.add_op(
                array_map(elem_opt_ty, self.length, unit_ty), array, func
            )
        return []

    def define_unwrap_none_helper(self) -> hf.Function:
        """Define an `unwrap_none` function that checks that the passed element is
        indeed `None`."""
        opt_ty = ht.Option(ht.Variable(0, ht.TypeBound.Any))
        unit_ty = ht.UnitSum(1)
        func_ty = ht.PolyFuncType(
            params=[ht.TypeTypeParam(ht.TypeBound.Any)],
            body=ht.FunctionType([opt_ty], [unit_ty]),
        )
        func, already_defined = self.ctx.declare_global_func(
            ARRAY_ITER_ASSERT_ALL_USED_HELPER, func_ty
        )
        if not already_defined:
            err_msg = "ArrayIter._assert_all_used: array element has not been used"
            build_expect_none(func, func.inputs()[0], err_msg)
            func.set_outputs(func.add_op(ops.MakeTuple()))
        return func



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")


@guppy(module)
def foo(x: float) -> int:
    return x


# Call check instead of compile
module.check()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit

module = GuppyModule("test")
module.load(qubit)


@guppy.declare(module)
def foo(q: qubit) -> None: ...


@guppy(module)
def test() -> None:
   foo(qubit())


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")


@guppy.struct(module)
class MyStruct:
    x: tuple[int, int]


@guppy(module)
def main() -> None:
    MyStruct(0)


module.compile()



import ast
import inspect
from collections.abc import Callable, KeysView, Sequence
from dataclasses import dataclass, field
from pathlib import Path
from types import ModuleType
from typing import Any, TypeVar, cast, overload

from hugr import ops
from hugr import tys as ht
from hugr import val as hv
from hugr.package import FuncDefnPointer, ModulePointer

import guppylang
from guppylang.ast_util import annotate_location
from guppylang.definition.common import DefId, Definition
from guppylang.definition.const import RawConstDef
from guppylang.definition.custom import (
    CustomCallChecker,
    CustomInoutCallCompiler,
    DefaultCallChecker,
    NotImplementedCallCompiler,
    OpCompiler,
    RawCustomFunctionDef,
)
from guppylang.definition.declaration import RawFunctionDecl
from guppylang.definition.extern import RawExternDef
from guppylang.definition.function import (
    CompiledFunctionDef,
    RawFunctionDef,
)
from guppylang.definition.parameter import ConstVarDef, TypeVarDef
from guppylang.definition.pytket_circuits import (
    RawLoadPytketDef,
    RawPytketDef,
)
from guppylang.definition.struct import RawStructDef
from guppylang.definition.ty import OpaqueTypeDef, TypeDef
from guppylang.error import MissingModuleError, pretty_errors
from guppylang.ipython_inspect import (
    get_ipython_globals,
    is_ipython_dummy_file,
    is_running_ipython,
)
from guppylang.module import (
    GuppyModule,
    PyClass,
    PyFunc,
    find_guppy_module_in_py_module,
    get_calling_frame,
    sphinx_running,
)
from guppylang.span import Loc, SourceMap, Span
from guppylang.tys.arg import Argument
from guppylang.tys.param import Parameter
from guppylang.tys.subst import Inst
from guppylang.tys.ty import NumericType

S = TypeVar("S")
T = TypeVar("T")
Decorator = Callable[[S], T]

FuncDefDecorator = Decorator[PyFunc, RawFunctionDef]
FuncDeclDecorator = Decorator[PyFunc, RawFunctionDecl]
CustomFuncDecorator = Decorator[PyFunc, RawCustomFunctionDef]
PytketDecorator = Decorator[PyFunc, RawPytketDef]
ClassDecorator = Decorator[PyClass, PyClass]
OpaqueTypeDecorator = Decorator[PyClass, OpaqueTypeDef]
StructDecorator = Decorator[PyClass, RawStructDef]


_JUPYTER_NOTEBOOK_MODULE = "<jupyter-notebook>"


@dataclass(frozen=True)
class ModuleIdentifier:
    """Identifier for the Python file/module that called the decorator."""

    filename: Path

    #: The name of the module. We only store this to have nice name to report back to
    #: the user. When determining whether two `ModuleIdentifier`s correspond to the same
    #: module, we only take the module path into account.
    name: str = field(compare=False)

    #: A reference to the python module
    module: ModuleType | None = field(compare=False)


class _Guppy:
    """Class for the `@guppy` decorator."""

    # The currently-alive GuppyModules, associated with a Python file/module
    _modules: dict[ModuleIdentifier, GuppyModule]

    # Storage for source code that has been read by the compiler
    _sources: SourceMap

    def __init__(self) -> None:
        self._modules = {}
        self._sources = SourceMap()

    @overload
    def __call__(self, arg: PyFunc) -> RawFunctionDef: ...

    @overload
    def __call__(self, arg: GuppyModule) -> FuncDefDecorator: ...

    @pretty_errors
    def __call__(self, arg: PyFunc | GuppyModule) -> FuncDefDecorator | RawFunctionDef:
        """Decorator to annotate Python functions as Guppy code.

        Optionally, the `GuppyModule` in which the function should be placed can
        be passed to the decorator.
        """

        def dec(f: Callable[..., Any], module: GuppyModule) -> RawFunctionDef:
            return module.register_func_def(f)

        return self._with_optional_module(dec, arg)

    @overload  # Always S != GuppyModule, hence ok to:
    def _with_optional_module(  # type: ignore[overload-overlap]
        self, dec: Callable[[S, GuppyModule], T], arg: S
    ) -> T: ...

    @overload
    def _with_optional_module(
        self, dec: Callable[[S, GuppyModule], T], arg: GuppyModule
    ) -> Decorator[S, T]: ...

    def _with_optional_module(
        self, dec: Callable[[S, GuppyModule], T], arg: S | GuppyModule
    ) -> Decorator[S, T] | T:
        """Helper function to define decorators that take an optional `GuppyModule`
        argument but no other arguments.

        For example, we allow `@guppy(module)` but also `@guppy`.
        """
        if isinstance(arg, GuppyModule):
            return lambda s: dec(s, arg)
        return dec(arg, self.get_module())

    def _get_python_caller(self, fn: PyFunc | None = None) -> ModuleIdentifier:
        """Returns an identifier for the Python file/module that called the decorator.

        :param fn: Optional. The function that was decorated.
        """
        if fn is not None:
            filename = inspect.getfile(fn)
            module = inspect.getmodule(fn)
        else:
            frame = inspect.currentframe()
            # loop to skip frames from the `pretty_error` decorator
            while frame:
                info = inspect.getframeinfo(frame)
                if info and info.filename != __file__:
                    module = inspect.getmodule(frame)
                    if module != guppylang.error:
                        break
                frame = frame.f_back
            else:
                raise RuntimeError("Could not find a caller for the `@guppy` decorator")

            # Jupyter notebook cells all get different dummy filenames. However,
            # we want the whole notebook to correspond to a single implicit
            # Guppy module.
            filename = info.filename
            if is_running_ipython() and not module and is_ipython_dummy_file(filename):
                filename = _JUPYTER_NOTEBOOK_MODULE
        module_path = Path(filename)
        return ModuleIdentifier(
            module_path, module.__name__ if module else module_path.name, module
        )

    def init_module(self, import_builtins: bool = True) -> None:
        """Manually initialises a Guppy module for the current Python file.

        Calling this method is only required when trying to define an empty module or
        a module that doesn't include the builtins.
        """
        module_id = self._get_python_caller()
        if module_id in self._modules:
            msg = f"Module {module_id.name} is already initialised"
            raise ValueError(msg)
        self._modules[module_id] = GuppyModule(module_id.name, import_builtins)

    @pretty_errors
    def extend_type(
        self, defn: TypeDef, module: GuppyModule | None = None
    ) -> ClassDecorator:
        """Decorator to add new instance functions to a type."""
        mod = module or self.get_module()
        mod._instance_func_buffer = {}

        def dec(c: type) -> type:
            mod._register_buffered_instance_funcs(defn)
            return c

        return dec

    @pretty_errors
    def type(
        self,
        hugr_ty: ht.Type | Callable[[Sequence[Argument]], ht.Type],
        name: str = "",
        copyable: bool = True,
        droppable: bool = True,
        bound: ht.TypeBound | None = None,
        params: Sequence[Parameter] | None = None,
        module: GuppyModule | None = None,
    ) -> OpaqueTypeDecorator:
        """Decorator to annotate a class definitions as Guppy types.

        Requires the static Hugr translation of the type. Additionally, the type can be
        marked as linear. All `@guppy` annotated functions on the class are turned into
        instance functions.

        For non-generic types, the Hugr representation can be passed as a static value.
        For generic types, a callable may be passed that takes the type arguments of a
        concrete instantiation.
        """
        mod = module or self.get_module()
        mod._instance_func_buffer = {}

        mk_hugr_ty = (lambda _: hugr_ty) if isinstance(hugr_ty, ht.Type) else hugr_ty

        def dec(c: type) -> OpaqueTypeDef:
            defn = OpaqueTypeDef(
                DefId.fresh(mod),
                name or c.__name__,
                None,
                params or [],
                not copyable,
                not droppable,
                mk_hugr_ty,
                bound,
            )
            mod.register_def(defn)
            mod._register_buffered_instance_funcs(defn)
            return defn

        return dec

    @property
    def struct(
        self,
    ) -> Callable[[PyClass | GuppyModule], StructDecorator | RawStructDef]:
        """Decorator to define a new struct."""
        # Note that this is a property. Thus, the code below is executed *before*
        # the members of the decorated class are executed.
        # At this point, we don't know if the user has called `@struct(module)` or
        # just `@struct`. To be safe, we initialise the method buffer of the implicit
        # module either way
        caller_id = self._get_python_caller()
        implicit_module_existed = caller_id in self._modules
        implicit_module = self.get_module(
            # But don't try to do implicit imports since we're not sure if this is
            # actually an implicit module
            resolve_implicit_imports=False
        )
        implicit_module._instance_func_buffer = {}

        # Extract Python scope from the frame that called `guppy.struct`
        frame = get_calling_frame()
        python_scope = frame.f_globals | frame.f_locals if frame else {}

        def dec(cls: type, module: GuppyModule) -> RawStructDef:
            defn = RawStructDef(
                DefId.fresh(module), cls.__name__, None, cls, python_scope
            )
            module.register_def(defn)
            module._register_buffered_instance_funcs(defn)
            # If we mistakenly initialised the method buffer of the implicit module
            # we can just clear it here
            if module != implicit_module:
                assert implicit_module._instance_func_buffer == {}
                implicit_module._instance_func_buffer = None
                if not implicit_module_existed:
                    self._modules.pop(caller_id)
            return defn

        def higher_dec(arg: GuppyModule | PyClass) -> StructDecorator | RawStructDef:
            if isinstance(arg, GuppyModule):
                arg._instance_func_buffer = {}
            return self._with_optional_module(dec, arg)

        return higher_dec

    @pretty_errors
    def type_var(
        self,
        name: str,
        copyable: bool = True,
        droppable: bool = True,
        module: GuppyModule | None = None,
    ) -> TypeVar:
        """Creates a new type variable in a module."""
        module = module or self.get_module()
        defn = TypeVarDef(DefId.fresh(module), name, None, copyable, droppable)
        module.register_def(defn)
        # Return an actual Python `TypeVar` so it can be used as an actual type in code
        # that is executed by interpreter before handing it to Guppy.
        return TypeVar(name)

    @pretty_errors
    def nat_var(self, name: str, module: GuppyModule | None = None) -> TypeVar:
        """Creates a new const nat variable in a module."""
        module = module or self.get_module()
        defn = ConstVarDef(
            DefId.fresh(module), name, None, NumericType(NumericType.Kind.Nat)
        )
        module.register_def(defn)
        # Return an actual Python `TypeVar` so it can be used as an actual type in code
        # that is executed by interpreter before handing it to Guppy.
        return TypeVar(name)

    @pretty_errors
    def custom(
        self,
        compiler: CustomInoutCallCompiler | None = None,
        checker: CustomCallChecker | None = None,
        higher_order_value: bool = True,
        name: str = "",
        module: GuppyModule | None = None,
    ) -> CustomFuncDecorator:
        """Decorator to add custom typing or compilation behaviour to function decls.

        Optionally, usage of the function as a higher-order value can be disabled. In
        that case, the function signature can be omitted if a custom call compiler is
        provided.
        """
        mod = module or self.get_module()

        def dec(f: PyFunc) -> RawCustomFunctionDef:
            call_checker = checker or DefaultCallChecker()
            func = RawCustomFunctionDef(
                DefId.fresh(mod),
                name or f.__name__,
                None,
                f,
                call_checker,
                compiler or NotImplementedCallCompiler(),
                higher_order_value,
            )
            mod.register_def(func)
            return func

        return dec

    def hugr_op(
        self,
        op: Callable[[ht.FunctionType, Inst], ops.DataflowOp],
        checker: CustomCallChecker | None = None,
        higher_order_value: bool = True,
        name: str = "",
        module: GuppyModule | None = None,
    ) -> CustomFuncDecorator:
        """Decorator to annotate function declarations as HUGR ops.

        Args:
            module: The module in which the function should be defined.
            op: A function that takes an instantiation of the type arguments as well as
                the inferred input and output types and returns a concrete HUGR op.
            checker: The custom call checker.
            higher_order_value: Whether the function may be used as a higher-order
                value.
            name: The name of the function.
        """
        return self.custom(OpCompiler(op), checker, higher_order_value, name, module)

    @overload
    def declare(self, arg: GuppyModule) -> RawFunctionDecl: ...

    @overload
    def declare(self, arg: PyFunc) -> FuncDeclDecorator: ...

    def declare(self, arg: GuppyModule | PyFunc) -> FuncDeclDecorator | RawFunctionDecl:
        """Decorator to declare functions"""

        def dec(f: Callable[..., Any], module: GuppyModule) -> RawFunctionDecl:
            return module.register_func_decl(f)

        return self._with_optional_module(dec, arg)

    def constant(
        self, name: str, ty: str, value: hv.Value, module: GuppyModule | None = None
    ) -> RawConstDef:
        """Adds a constant to a module, backed by a `hugr.val.Value`."""
        module = module or self.get_module()
        type_ast = _parse_expr_string(
            ty, f"Not a valid Guppy type: `{ty}`", self._sources
        )
        defn = RawConstDef(DefId.fresh(module), name, None, type_ast, value)
        module.register_def(defn)
        return defn

    def extern(
        self,
        name: str,
        ty: str,
        symbol: str | None = None,
        constant: bool = True,
        module: GuppyModule | None = None,
    ) -> RawExternDef:
        """Adds an extern symbol to a module."""
        module = module or self.get_module()
        type_ast = _parse_expr_string(
            ty, f"Not a valid Guppy type: `{ty}`", self._sources
        )
        defn = RawExternDef(
            DefId.fresh(module), name, None, symbol or name, constant, type_ast
        )
        module.register_def(defn)
        return defn

    def load(self, m: ModuleType | GuppyModule) -> None:
        caller = self._get_python_caller()
        if caller not in self._modules:
            self._modules[caller] = GuppyModule(caller.name)
        module = self._modules[caller]
        module.load_all(m)

    def get_module(
        self, id: ModuleIdentifier | None = None, resolve_implicit_imports: bool = True
    ) -> GuppyModule:
        """Returns the local GuppyModule."""
        if id is None:
            id = self._get_python_caller()
        if id not in self._modules:
            self._modules[id] = GuppyModule(id.name.split(".")[-1])
        module = self._modules[id]
        # Update implicit imports
        if resolve_implicit_imports:
            globs: dict[str, Any] = {}
            if id.module:
                globs = id.module.__dict__
            # Jupyter notebooks are not made up of a single module, so we need to find
            # it's globals by querying the ipython kernel
            elif id.name == _JUPYTER_NOTEBOOK_MODULE:
                globs = get_ipython_globals()
            if globs:
                defs: dict[str, Definition | ModuleType] = {}
                for x, value in globs.items():
                    if isinstance(value, Definition):
                        other_module = value.id.module
                        if other_module and other_module != module:
                            defs[x] = value
                    elif isinstance(value, ModuleType):
                        try:
                            other_module = find_guppy_module_in_py_module(value)
                            if other_module and other_module != module:
                                defs[x] = value
                        except ValueError:
                            pass
                module.load(**defs)
        return module

    def compile_module(self, id: ModuleIdentifier | None = None) -> ModulePointer:
        """Compiles the xlocal module into a Hugr."""
        module = self.get_module(id)
        if not module:
            err = (
                f"Module {id.name} not found."
                if id
                else "No Guppy functions or types defined in this module."
            )
            raise MissingModuleError(err)
        return module.compile()

    def compile_function(self, f_def: RawFunctionDef) -> FuncDefnPointer:
        """Compiles a single function definition."""
        module = f_def.id.module
        if not module:
            raise ValueError("Function definition must belong to a module")
        compiled_module = module.compile()
        assert module._compiled is not None, "Module should be compiled"
        globs = module._compiled.context
        assert globs is not None
        compiled_def = globs.build_compiled_def(f_def.id)
        assert isinstance(compiled_def, CompiledFunctionDef)
        node = compiled_def.func_def.parent_node
        return FuncDefnPointer(
            compiled_module.package, compiled_module.module_index, node
        )

    def registered_modules(self) -> KeysView[ModuleIdentifier]:
        """Returns a list of all currently registered modules for local contexts."""
        return self._modules.keys()

    @pretty_errors
    def pytket(
        self, input_circuit: Any, module: GuppyModule | None = None
    ) -> PytketDecorator:
        """Adds a pytket circuit function definition with explicit signature."""
        err_msg = "Only pytket circuits can be passed to guppy.pytket"
        try:
            import pytket

            if not isinstance(input_circuit, pytket.circuit.Circuit):
                raise TypeError(err_msg) from None

        except ImportError:
            raise TypeError(err_msg) from None

        mod = module or self.get_module()

        def func(f: PyFunc) -> RawPytketDef:
            return mod.register_pytket_func(f, input_circuit)

        return func

    @pretty_errors
    def load_pytket(
        self,
        name: str,
        input_circuit: Any,
        module: GuppyModule | None = None,
        *,
        use_arrays: bool = True,
    ) -> RawLoadPytketDef:
        """Adds a pytket circuit function definition with implicit signature."""
        err_msg = "Only pytket circuits can be passed to guppy.load_pytket"
        try:
            import pytket

            if not isinstance(input_circuit, pytket.circuit.Circuit):
                raise TypeError(err_msg) from None

        except ImportError:
            raise TypeError(err_msg) from None

        mod = module or self.get_module()
        span = _find_load_call(self._sources)
        defn = RawLoadPytketDef(
            DefId.fresh(module), name, None, span, input_circuit, use_arrays
        )
        mod.register_def(defn)
        return defn


class _GuppyDummy:
    """A dummy class with the same interface as `@guppy` that is used during sphinx
    builds to mock the decorator.
    """

    _sources = SourceMap()

    def __call__(self, arg: PyFunc | GuppyModule) -> Any:
        if isinstance(arg, GuppyModule):
            return lambda f: f
        return arg

    def init_module(self, *args: Any, **kwargs: Any) -> None:
        pass

    def extend_type(self, *args: Any, **kwargs: Any) -> Any:
        return lambda cls: cls

    def type(self, *args: Any, **kwargs: Any) -> Any:
        return lambda cls: cls

    def struct(self, *args: Any, **kwargs: Any) -> Any:
        return lambda cls: cls

    def type_var(self, name: str, *args: Any, **kwargs: Any) -> Any:
        # Return an actual type variable so it can be used in `Generic[...]`
        return TypeVar(name)

    def nat_var(self, name: str, *args: Any, **kwargs: Any) -> Any:
        # Return an actual type variable so it can be used in `Generic[...]`
        return TypeVar(name)

    def custom(self, *args: Any, **kwargs: Any) -> Any:
        return lambda f: f

    def hugr_op(self, *args: Any, **kwargs: Any) -> Any:
        return lambda f: f

    def declare(self, arg: PyFunc | GuppyModule) -> Any:
        if isinstance(arg, GuppyModule):
            return lambda f: f
        return arg

    def constant(self, *args: Any, **kwargs: Any) -> Any:
        return None

    def extern(self, *args: Any, **kwargs: Any) -> Any:
        return None

    def load(self, *args: Any, **kwargs: Any) -> None:
        pass

    def get_module(self, *args: Any, **kwargs: Any) -> Any:
        return GuppyModule("dummy", import_builtins=False)


guppy = cast(_Guppy, _GuppyDummy()) if sphinx_running() else _Guppy()


def _parse_expr_string(ty_str: str, parse_err: str, sources: SourceMap) -> ast.expr:
    """Helper function to parse expressions that are provided as strings.

    Tries to infer the source location were the given string was defined by inspecting
    the call stack.
    """
    try:
        expr_ast = ast.parse(ty_str, mode="eval").body
    except SyntaxError:
        raise SyntaxError(parse_err) from None

    # Try to annotate the type AST with source information. This requires us to
    # inspect the stack frame of the caller
    if caller_frame := get_calling_frame():
        info = inspect.getframeinfo(caller_frame)
        if caller_module := inspect.getmodule(caller_frame):
            sources.add_file(info.filename)
            source_lines, _ = inspect.getsourcelines(caller_module)
            source = "".join(source_lines)
            annotate_location(expr_ast, source, info.filename, 1)
            # Modify the AST so that all sub-nodes span the entire line. We
            # can't give a better location since we don't know the column
            # offset of the `ty` argument
            for node in [expr_ast, *ast.walk(expr_ast)]:
                node.lineno = node.end_lineno = info.lineno
                node.col_offset = 0
                node.end_col_offset = len(source_lines[info.lineno - 1]) - 1
    return expr_ast


def _find_load_call(sources: SourceMap) -> Span | None:
    """Helper function to find location where pytket circuit was loaded.

    Tries to define a source code span by inspecting the call stack.
    """
    # Go back as first frame outside of compiler modules is 'pretty_errors_wrapped'.
    if (caller_frame := get_calling_frame()) and (load_frame := caller_frame.f_back):
        info = inspect.getframeinfo(load_frame)
        filename = info.filename
        lineno = info.lineno
        sources.add_file(filename)
        # If we don't support python <= 3.10, this can be done better with
        # info.positions which gives you exact offsets.
        # For now over approximate and make the span cover the entire line.
        if load_module := inspect.getmodule(load_frame):
            source_lines, _ = inspect.getsourcelines(load_module)
            max_offset = len(source_lines[lineno - 1]) - 1

            start = Loc(filename, lineno, 0)
            end = Loc(filename, lineno, max_offset)
            return Span(start, end)
    return None



from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Literal, TypeGuard

import hugr.std
import hugr.std.collections.array
import hugr.std.collections.list
from hugr import tys as ht

from guppylang.ast_util import AstNode
from guppylang.definition.common import DefId
from guppylang.definition.ty import OpaqueTypeDef, TypeDef
from guppylang.error import GuppyError, InternalGuppyError
from guppylang.experimental import check_lists_enabled
from guppylang.tys.arg import Argument, ConstArg, TypeArg
from guppylang.tys.const import Const, ConstValue
from guppylang.tys.errors import WrongNumberOfTypeArgsError
from guppylang.tys.param import ConstParam, TypeParam
from guppylang.tys.ty import (
    FunctionType,
    NoneType,
    NumericType,
    OpaqueType,
    TupleType,
    Type,
)

if TYPE_CHECKING:
    from guppylang.checker.core import Globals


@dataclass(frozen=True)
class CallableTypeDef(TypeDef):
    """Type definition associated with the builtin `Callable` type.

    Any impls on functions can be registered with this definition.
    """

    name: Literal["Callable"] = field(default="Callable", init=False)

    def check_instantiate(
        self, args: Sequence[Argument], globals: "Globals", loc: AstNode | None = None
    ) -> FunctionType:
        # Callable types are constructed using special logic in the type parser
        raise InternalGuppyError("Tried to `Callable` type via `check_instantiate`")


@dataclass(frozen=True)
class _TupleTypeDef(TypeDef):
    """Type definition associated with the builtin `tuple` type.

    Any impls on tuples can be registered with this definition.
    """

    name: Literal["tuple"] = field(default="tuple", init=False)

    def check_instantiate(
        self, args: Sequence[Argument], globals: "Globals", loc: AstNode | None = None
    ) -> TupleType:
        # We accept any number of arguments. If users just write `tuple`, we give them
        # the empty tuple type. We just have to make sure that the args are of kind type
        args = [
            # TODO: Better error location
            TypeParam(0, f"T{i}", must_be_copyable=False, must_be_droppable=False)
            .check_arg(arg, loc)
            .ty
            for i, arg in enumerate(args)
        ]
        return TupleType(args)


@dataclass(frozen=True)
class _NoneTypeDef(TypeDef):
    """Type definition associated with the builtin `None` type.

    Any impls on None can be registered with this definition.
    """

    name: Literal["None"] = field(default="None", init=False)

    def check_instantiate(
        self, args: Sequence[Argument], globals: "Globals", loc: AstNode | None = None
    ) -> NoneType:
        if args:
            raise GuppyError(WrongNumberOfTypeArgsError(loc, 0, len(args), "None"))
        return NoneType()


@dataclass(frozen=True)
class _NumericTypeDef(TypeDef):
    """Type definition associated with the builtin numeric types.

    Any impls on numerics can be registered with these definitions.
    """

    ty: NumericType

    def check_instantiate(
        self, args: Sequence[Argument], globals: "Globals", loc: AstNode | None = None
    ) -> NumericType:
        if args:
            raise GuppyError(WrongNumberOfTypeArgsError(loc, 0, len(args), self.name))
        return self.ty


@dataclass(frozen=True)
class _ListTypeDef(OpaqueTypeDef):
    """Type definition associated with the builtin `list` type.

    We have a custom definition to disable usage of lists unless experimental features
    are enabled.
    """

    def check_instantiate(
        self, args: Sequence[Argument], globals: "Globals", loc: AstNode | None = None
    ) -> OpaqueType:
        check_lists_enabled(loc)
        return super().check_instantiate(args, globals, loc)


def _list_to_hugr(args: Sequence[Argument]) -> ht.Type:
    # Type checker ensures that we get a single arg of kind type
    [arg] = args
    assert isinstance(arg, TypeArg)
    # Linear elements are turned into an optional to enable unsafe indexing.
    # See `ListGetitemCompiler` for details.
    elem_ty = ht.Option(arg.ty.to_hugr()) if arg.ty.linear else arg.ty.to_hugr()
    return hugr.std.collections.list.List(elem_ty)


def _array_to_hugr(args: Sequence[Argument]) -> ht.Type:
    # Type checker ensures that we get a two args
    [ty_arg, len_arg] = args
    assert isinstance(ty_arg, TypeArg)
    assert isinstance(len_arg, ConstArg)

    # Linear elements are turned into an optional to enable unsafe indexing.
    # See `ArrayGetitemCompiler` for details.
    # Same also for classical arrays, see https://github.com/CQCL/guppylang/issues/629
    elem_ty = ht.Option(ty_arg.ty.to_hugr())
    hugr_arg = len_arg.to_hugr()

    return hugr.std.collections.array.Array(elem_ty, hugr_arg)


def _sized_iter_to_hugr(args: Sequence[Argument]) -> ht.Type:
    [ty_arg, len_arg] = args
    assert isinstance(ty_arg, TypeArg)
    assert isinstance(len_arg, ConstArg)
    return ty_arg.ty.to_hugr()


def _option_to_hugr(args: Sequence[Argument]) -> ht.Type:
    [arg] = args
    assert isinstance(arg, TypeArg)
    return ht.Option(arg.ty.to_hugr())


callable_type_def = CallableTypeDef(DefId.fresh(), None)
tuple_type_def = _TupleTypeDef(DefId.fresh(), None)
none_type_def = _NoneTypeDef(DefId.fresh(), None)
bool_type_def = OpaqueTypeDef(
    id=DefId.fresh(),
    name="bool",
    defined_at=None,
    params=[],
    never_copyable=False,
    never_droppable=False,
    to_hugr=lambda _: ht.Bool,
)
nat_type_def = _NumericTypeDef(
    DefId.fresh(), "nat", None, NumericType(NumericType.Kind.Nat)
)
int_type_def = _NumericTypeDef(
    DefId.fresh(), "int", None, NumericType(NumericType.Kind.Int)
)
float_type_def = _NumericTypeDef(
    DefId.fresh(), "float", None, NumericType(NumericType.Kind.Float)
)
string_type_def = OpaqueTypeDef(
    id=DefId.fresh(),
    name="str",
    defined_at=None,
    params=[],
    never_copyable=False,
    never_droppable=False,
    to_hugr=lambda _: hugr.std.PRELUDE.get_type("string").instantiate([]),
)
list_type_def = _ListTypeDef(
    id=DefId.fresh(),
    name="list",
    defined_at=None,
    params=[TypeParam(0, "T", must_be_copyable=False, must_be_droppable=False)],
    never_copyable=False,
    never_droppable=False,
    to_hugr=_list_to_hugr,
)
array_type_def = OpaqueTypeDef(
    id=DefId.fresh(),
    name="array",
    defined_at=None,
    params=[
        TypeParam(0, "T", must_be_copyable=False, must_be_droppable=False),
        ConstParam(1, "n", NumericType(NumericType.Kind.Nat)),
    ],
    never_copyable=True,
    never_droppable=False,
    to_hugr=_array_to_hugr,
)
sized_iter_type_def = OpaqueTypeDef(
    id=DefId.fresh(),
    name="SizedIter",
    defined_at=None,
    params=[
        TypeParam(0, "T", must_be_copyable=False, must_be_droppable=False),
        ConstParam(1, "n", NumericType(NumericType.Kind.Nat)),
    ],
    never_copyable=False,
    never_droppable=False,
    to_hugr=_sized_iter_to_hugr,
)
option_type_def = OpaqueTypeDef(
    id=DefId.fresh(),
    name="Option",
    defined_at=None,
    params=[TypeParam(0, "T", must_be_copyable=False, must_be_droppable=False)],
    never_copyable=False,
    never_droppable=False,
    to_hugr=_option_to_hugr,
)


def bool_type() -> OpaqueType:
    return OpaqueType([], bool_type_def)


def nat_type() -> NumericType:
    return NumericType(NumericType.Kind.Nat)


def int_type() -> NumericType:
    return NumericType(NumericType.Kind.Int)


def float_type() -> NumericType:
    return NumericType(NumericType.Kind.Float)


def string_type() -> OpaqueType:
    return OpaqueType([], string_type_def)


def list_type(element_ty: Type) -> OpaqueType:
    return OpaqueType([TypeArg(element_ty)], list_type_def)


def array_type(element_ty: Type, length: int | Const) -> OpaqueType:
    if isinstance(length, int):
        length = ConstValue(nat_type(), length)
    return OpaqueType([TypeArg(element_ty), ConstArg(length)], array_type_def)


def sized_iter_type(iter_type: Type, size: int | Const) -> OpaqueType:
    if isinstance(size, int):
        size = ConstValue(nat_type(), size)
    return OpaqueType([TypeArg(iter_type), ConstArg(size)], sized_iter_type_def)


def option_type(element_ty: Type) -> OpaqueType:
    return OpaqueType([TypeArg(element_ty)], option_type_def)


def is_bool_type(ty: Type) -> bool:
    return isinstance(ty, OpaqueType) and ty.defn == bool_type_def


def is_string_type(ty: Type) -> bool:
    return isinstance(ty, OpaqueType) and ty.defn == string_type_def


def is_list_type(ty: Type) -> bool:
    return isinstance(ty, OpaqueType) and ty.defn == list_type_def


def is_array_type(ty: Type) -> TypeGuard[OpaqueType]:
    return isinstance(ty, OpaqueType) and ty.defn == array_type_def


def is_sized_iter_type(ty: Type) -> TypeGuard[OpaqueType]:
    return isinstance(ty, OpaqueType) and ty.defn == sized_iter_type_def


def get_element_type(ty: Type) -> Type:
    assert isinstance(ty, OpaqueType)
    assert ty.defn in (list_type_def, array_type_def, option_type_def)
    (arg, *_) = ty.args
    assert isinstance(arg, TypeArg)
    return arg.ty


def get_iter_size(ty: Type) -> Const:
    assert isinstance(ty, OpaqueType)
    assert ty.defn == sized_iter_type_def
    match ty.args:
        case [_, ConstArg(const)]:
            return const
        case _:
            raise InternalGuppyError("Unexpected type args")





from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit

import guppylang.std.quantum as quantum


def test_basic(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy(module)
    def test(q: qubit @ owned) -> qubit:
        foo(q)
        return q

    validate(module.compile())


def test_mixed(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q1: qubit, q2: qubit @ owned) -> qubit: ...

    @guppy(module)
    def test(q1: qubit @ owned, q2: qubit @ owned) -> tuple[qubit, qubit]:
        q2 = foo(q1, q2)
        return q1, q2

    validate(module.compile())


def test_local(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy(module)
    def test(q: qubit @ owned) -> qubit:
        f = foo
        f(q)
        return q

    validate(module.compile())


def test_nested_calls(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(x: int, q: qubit) -> int: ...

    @guppy(module)
    def test(q: qubit @ owned) -> tuple[int, qubit]:
        # This is legal since function arguments and tuples are evaluated left to right
        return foo(foo(foo(0, q), q), q), q

    validate(module.compile())


def test_struct(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.struct(module)
    class MyStruct:
        q1: qubit
        q2: qubit

    @guppy.declare(module)
    def foo(q1: qubit, q2: qubit) -> None: ...

    @guppy.declare(module)
    def bar(a: MyStruct) -> None: ...

    @guppy(module)
    def test1(a: MyStruct @ owned) -> MyStruct:
        foo(a.q1, a.q2)
        bar(a)
        return a

    @guppy(module)
    def test2(a: MyStruct @ owned) -> MyStruct:
        bar(a)
        foo(a.q1, a.q2)
        bar(a)
        return a

    validate(module.compile())


def test_control_flow(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy.declare(module)
    def bar(q: qubit) -> bool: ...

    @guppy(module)
    def test(q1: qubit @ owned, q2: qubit @ owned, n: int) -> tuple[qubit, qubit]:
        i = 0
        while i < n:
            foo(q1)
            if bar(q1) or bar(q2):
                foo(q2)
                continue
            elif not bar(q2):
                return q1, q2
            foo(q1)
            if bar(q2):
                foo(q1)
                break
            else:
                foo(q2)
            foo(q2)
            foo(q1)
            if bar(q2) and bar(q1):
                if i > 5:
                    foo(q2)
                return q1, q2
            foo(q1)
            i += 1
        return q1, q2

    validate(module.compile())


def test_tensor(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.struct(module)
    class A:
        q: qubit

    @guppy.struct(module)
    class B:
        q: qubit
        x: int

    @guppy.struct(module)
    class C:
        q: qubit
        x: float

    @guppy.declare(module)
    def foo(a: A, x: int) -> None: ...

    @guppy.declare(module)
    def bar(y: float, b: B, c: C @ owned) -> C: ...

    @guppy.declare(module)
    def baz(c: C) -> None: ...

    @guppy(module)
    def test(
        a: A @ owned, b: B @ owned, c1: C @ owned, c2: C @ owned, x: bool
    ) -> tuple[A, B, C, C]:
        c1 = (foo, bar, baz)(a, b.x, c1.x, b, c1, c2)
        if x:
            c1 = ((foo, bar), baz)(a, b.x, c1.x, b, c1, c2)
        c1 = (foo, (bar, baz))(a, b.x, c1.x, b, c1, c2)
        return a, b, c1, c2

    validate(module.compile())


def test_basic_def(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def h(q: qubit) -> None: ...

    @guppy(module)
    def foo(q: qubit) -> None:
        h(q)
        h(q)

    @guppy(module)
    def test(q: qubit @ owned) -> qubit:
        foo(q)
        foo(q)
        return q

    validate(module.compile())


def test_empty_def(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy(module)
    def test(q: qubit) -> None:
        pass

    @guppy(module)
    def main(q: qubit @ owned) -> qubit:
        test(q)
        return q

    validate(module.compile())


def test_mixed_def(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy(module)
    def test(
        b: int, c: qubit, d: float, a: tuple[qubit, qubit], e: qubit @ owned
    ) -> tuple[qubit, float]:
        foo(c)
        return e, b + d

    validate(module.compile())


def test_move_back(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.struct(module)
    class MyStruct:
        q: qubit

    @guppy.declare(module)
    def use(q: qubit @ owned) -> None: ...

    @guppy(module)
    def foo(s: MyStruct) -> None:
        use(s.q)
        s.q = qubit()

    @guppy(module)
    def bar(s: MyStruct) -> None:
        s.q = s.q

    @guppy(module)
    def swap(s: MyStruct, t: MyStruct) -> None:
        s.q, t.q = t.q, s.q

    @guppy(module)
    def main(s: MyStruct @ owned, t: MyStruct @ owned) -> MyStruct:
        foo(s)
        swap(s, t)
        bar(t)
        use(t.q)
        return s

    validate(module.compile())


def test_move_back_branch(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.struct(module)
    class MyStruct:
        q: qubit

    @guppy.declare(module)
    def use(q: qubit @ owned) -> None: ...

    @guppy(module)
    def test(
        s: MyStruct, b: bool, n: int, q1: qubit @ owned, q2: qubit @ owned
    ) -> None:
        use(s.q)
        if b:
            s.q = q1
            use(q2)
        else:
            s.q = q2
            use(q1)
        use(s.q)
        i = 0
        while True:
            if i == n:
                s.q = qubit()
                return
            i += 1

    @guppy(module)
    def main(s: MyStruct @ owned) -> MyStruct:
        test(s, False, 5, qubit(), qubit())
        return s

    validate(module.compile())


def test_self(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy.struct(module)
    class MyStruct:
        q: qubit

        @guppy(module)
        def bar(self: "MyStruct", b: bool) -> None:
            foo(self.q)
            if b:
                foo(self.q)

    validate(module.compile())


def test_subtype(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy(module)
    def main() -> qubit:
        q = qubit()
        foo(q)
        return q

    validate(module.compile())


def test_shadow_check(validate):
    module = GuppyModule("test")

    module.load(quantum, qubit)

    @guppy.declare(module)
    def foo(i: qubit) -> None: ...

    @guppy(module)
    def main(i: qubit) -> None:
        if True:
            foo(i)

    validate(module.compile())


def test_self_qubit(validate):
    module = GuppyModule("test")
    module.load(qubit)

    @guppy(module)
    def test() -> bool:
        q0 = qubit()

        result = q0.project_z()
        q0.measure()
        qubit().discard()
        return result

    validate(module.compile())


def test_non_terminating(validate):
    module = GuppyModule("test")
    module.load_all(quantum)

    @guppy.struct(module)
    class MyStruct:
        q1: qubit
        q2: qubit
        x: int

    @guppy.declare(module)
    def foo(q: qubit) -> None: ...

    @guppy.declare(module)
    def bar(s: MyStruct) -> None: ...

    @guppy(module)
    def test1(b: bool) -> None:
        q = qubit()
        s = MyStruct(qubit(), qubit(), 0)
        while True:
            foo(q)
            bar(s)

    @guppy(module)
    def test2(q: qubit, s: MyStruct, b: bool) -> None:
        while True:
            foo(q)
            if b:
                bar(s)

    @guppy(module)
    def test3(q: qubit, s: MyStruct) -> None:
        while True:
            pass

    validate(module.compile())



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")


@guppy.struct(module)
class MyStruct:
    x: int

    @guppy(module)
    def x(self: "MyStruct") -> int:
        return 0


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit

import guppylang.std.quantum as quantum


module = GuppyModule("test")

guppy.extern("x", ty="float[int]", module=module)

module.compile()
